from ._cli_common import main
from . import _cli_save_tasks
from . import _cli_run
from . import _cli_publish_tasks
from . import _cli_generate_cache
from . import _cli_generate_mosaic
from . import _cli_locate_grids



"""
USGS Landsat pixel quality

pq_band = input band for cloud masking
| Name | Units | Conversion | Description |
|------|-------|------------|-------------|
| QA_PIXEL | Bit Index | NA | Pixel quality; Bit: 0 = nodata; 1 = Dilated Cloud; 3 = cloud; 4 = cloud-shadow |

aerosol_band = input band for aerosol masking; provide one of the band as an input measurements
| Name | Units | Conversion | Description |
|------|-------|------------|-------------|
| SR_ATMOS_OPACITY | Unitless | 0.001 * DN | Atmospheric opacity; < 0.1 = clear; 0.1 - 0.3 = average; > 0.3 = hazy |
| SR_QA_AEROSOL    | Bit Index | NA | Aerosol level; Bit(6-7): 00 = climatology; 01 = low; 10 = medium; 11 = high |

filters = dict containing band-name as key and list of iterable tuples of morphological operations as it's value.
    Provide morphological operations in the order you want them to perform.
    For example,

    filters = {
        "clear_10_2_2": [("closing", 10), ("opening", 2), ("dilation", 2)]
    }

aerosol_filters = dict containing band-name as key and list of iterable tuples of morphological operations as it's value.
    Provide morphological operations in the order you want them to perform.
    Similar to filters but for aerosol.
    For example,

    aerosol_filters = {
        "clear_10_2_2_aerosol": [("closing", 10), ("opening", 2), ("dilation", 2)]
    }

resampling = "nearest"
"""

from functools import partial
from typing import Any, Dict, Optional, Tuple, Iterable

import dask.array as da
import xarray as xr
from datacube.utils import masking
from odc.algo import mask_cleanup, keep_good_only
from odc.algo._masking import _xr_fuse, _first_valid_np, _fuse_or_np

from odc.stats.model import StatsPluginInterface
from ._registry import register


class StatsPQLSBitmask(StatsPluginInterface):
    NAME = "pc_ls_bitmask"
    SHORT_NAME = NAME
    VERSION = '0.0.1'
    PRODUCT_FAMILY = "pixel_quality_statistics"

    def __init__(
            self,
            pq_band: str = "QA_PIXEL",
            aerosol_band: Optional[str] = None,
            # provide flags with high cloud bits definition
            flags: Dict[str, Optional[Any]] = dict(
                cloud="high_confidence",
                cirrus="high_confidence",
            ),
            nodata_flags: Dict[str, Optional[Any]] = dict(nodata=False),
            filters: Optional[Dict[str, Iterable[Tuple[str, int]]]] = None,
            aerosol_filters: Optional[Dict[str, Iterable[Tuple[str, int]]]] = None,
            resampling: str = "nearest",
            **kwargs
    ):
        self.pq_band = pq_band
        self.aerosol_band = aerosol_band
        input_bands = [self.pq_band]
        if self.aerosol_band:
            input_bands.append(self.aerosol_band)
        super().__init__(
                         input_bands=input_bands, resampling=resampling,
                         **kwargs)
        self.flags = flags
        self.nodata_flags = nodata_flags
        self.filters = filters or {}
        self.aerosol_filters = aerosol_filters or {}

    @property
    def measurements(self) -> Tuple[str, ...]:
        """
        Output product measurements
        """
        measurements = [
            "total",
            "clear",
            *list(self.filters)
        ]

        if self.aerosol_band:
            measurements.append("clear_aerosol")
            if self.aerosol_band == "SR_QA_AEROSOL":
                measurements.extend(list(self.aerosol_filters))

        return tuple(measurements)

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        """
        calculate pixel count:
        pq bands:
            total                  -> total pixel count (valid data)
            clear                  -> count clear_cloud (pixels without cloud)
            clear_<filter>         -> apply filter on erased_mask (cloud mask) and then count clear_cloud
            clear_aerosol          -> count clear_cloud + clear_aerosol
            clear_<filter>_aerosol -> count clear_cloud_filter + clear_aerosol
        """
        pq = xr.Dataset()

        for band, mask_filters in self.filters.items():
            erased_filter_band_name = band.replace("clear", "erased")
            xx[erased_filter_band_name] = mask_cleanup(xx["erased"], mask_filters=mask_filters)

        erased_bands = [str(n) for n in xx.data_vars if str(n).startswith("erased")]
        valid = xx["keeps"]
        pq["total"] = valid.sum(axis=0, dtype="uint16")
        for band in erased_bands:
            clear_band_name = band.replace("erased", "clear")
            if "aerosol" in band:
                pq[clear_band_name] = (valid & (~xx[band] & ~xx["erased"])).sum(axis=0, dtype="uint16")
            else:
                pq[clear_band_name] = (valid & (~xx[band])).sum(axis=0, dtype="uint16")

        if self.aerosol_band and self.aerosol_band == "SR_QA_AEROSOL":
            for band, mask_filters in self.aerosol_filters.items():
                erased_aerosol_filter_band_name = band.replace("clear", "erased")
                if erased_aerosol_filter_band_name not in xx:
                    xx[erased_aerosol_filter_band_name] = mask_cleanup(xx["erased"], mask_filters=mask_filters)

                pq[band] = (valid & (~xx[erased_aerosol_filter_band_name] & ~xx["erased_aerosol"])).sum(axis=0, dtype="uint16")

        return pq

    def native_transform(self, xx: xr.Dataset) -> xr.Dataset:
        """
        Loads the data in the native projection and perform transform
        bands:
            keeps          -> anything but nodata (valid pixels)
            erased         -> cloudy pixels
            erased_aerosol -> high aerosol pixels
        """
        pq_band = xx[self.pq_band]
        xx = xx.drop_vars([self.pq_band])

        flags_def = masking.get_flags_def(pq_band)

        # set cloud_mask - True=cloud, False=non-cloud
        mask, _ = masking.create_mask_value(flags_def, **self.flags)
        cloud_mask = (pq_band & mask) != 0

        # set no_data bitmask - True=data, False=no-data
        nodata_mask, _ = masking.create_mask_value(flags_def, **self.nodata_flags)
        keeps = (pq_band & nodata_mask) == 0

        if self.aerosol_band:
            aerosol_band = xx[self.aerosol_band]
            xx = xx.drop_vars([self.aerosol_band])
            # calculate aerosol_level or atmospheric opacity
            if self.aerosol_band == "SR_QA_AEROSOL":
                aerosol_level = da.bitwise_and(aerosol_band, 0b1100_0000) / 64
            elif self.aerosol_band == "SR_ATMOS_OPACITY":
                opacity = (aerosol_band.where(aerosol_band != -9999) * 0.001)

        # drops nodata pixels
        xx = keep_good_only(xx, keeps)

        xx["keeps"] = keeps
        xx["erased"] = cloud_mask
        if self.aerosol_band:
            if self.aerosol_band == "SR_QA_AEROSOL":
                xx["erased_aerosol"] = aerosol_level == 3
            elif self.aerosol_band == "SR_ATMOS_OPACITY":
                xx["erased_aerosol"] = opacity > 0.3

        return xx

    def fuser(self, xx: xr.Dataset) -> xr.Dataset:
        """
        Fuser cloud and aerosol masking bands with OR
        """
        cloud_mask = xx["erased"]
        xx = xx.drop_vars(["erased"])
        if self.aerosol_band:
            high_aerosol_mask = xx["erased_aerosol"]
            xx = xx.drop_vars(["erased_aerosol"])

        fuser_result = _xr_fuse(xx, partial(_first_valid_np, nodata=0), '')
        fuser_result["erased"] = _xr_fuse(cloud_mask, _fuse_or_np, cloud_mask.name)
        if self.aerosol_band:
            fuser_result["erased_aerosol"] = _xr_fuse(high_aerosol_mask, _fuse_or_np, high_aerosol_mask.name)

        return fuser_result


register("pq-ls-bitmask", StatsPQLSBitmask)



"""
Mangroves canopy cover classes
"""
from functools import partial
from itertools import product
from typing import Optional, Tuple

import numpy as np
import xarray as xr
import dask
import os
from odc.algo import keep_good_only, erase_bad
from odc.algo._masking import _fuse_mean_np, _fuse_or_np, _or_fuser, _xr_fuse
from odc.algo._percentile import xr_quantile_bands
from osgeo import gdal, ogr, osr

from ._registry import StatsPluginInterface, register

NODATA = 255


class Mangroves(StatsPluginInterface):

    NAME = "mangroves"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "mangroves"

    def __init__(
        self,
        pv_thresholds=[14, 38, 60],
        tcw_threshold=-1850,
        **kwargs,
    ):
        self.mangroves_extent = kwargs.pop('mangroves_extent', None)
        self.pv_thresholds = pv_thresholds
        self.tcw_threshold = tcw_threshold
        super().__init__(input_bands=["pv_pc_10", "qa", "wet_pc_10"], **kwargs)

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = [
            "canopy_cover_class"
        ]
        return _measurements
    
    def rasterize_mangroves_extent(self, shape_file, array_shape, orig_coords, resolution=(30, -30)):
        source_ds = ogr.Open(shape_file)
        source_layer = source_ds.GetLayer()

        yt, xt = array_shape[1:]
        xres, yres = resolution 
        no_data = 0
        xcoord, ycoord = orig_coords

        geotransform = (xcoord - (xres * 0.5), xres, 0, ycoord - (yres * 0.5), 0, yres)

        target_ds = gdal.GetDriverByName("MEM").Create("", xt, yt, gdal.GDT_Byte)
        target_ds.SetGeoTransform(geotransform)
        albers = osr.SpatialReference()
        albers.ImportFromEPSG(3577)
        target_ds.SetProjection(albers.ExportToWkt())
        band = target_ds.GetRasterBand(1)
        band.SetNoDataValue(no_data)

        gdal.RasterizeLayer(target_ds, [1], source_layer, burn_values=[1])
        return dask.array.from_array(band.ReadAsArray().reshape(array_shape), name=False)

    def fuser(self, xx):
        """
            no fuse required for mangroves since group by none
            return loaded data
        """
        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        """
            mangroves computation here
            it is not a 'reduce' though
        """
        if self.mangroves_extent:
            if not os.path.exists(self.mangroves_extent):
                raise FileNotFoundError(f"{self.mangroves_extent} not found")
            extent_mask = self.rasterize_mangroves_extent(self.mangroves_extent, xx.pv_pc_10.shape,
                                                     (xx.coords["x"].min(), xx.coords["y"].max()))
        else:
            extent_mask = dask.array.ones(xx.pv_pc_10.shape)
        good_data = (extent_mask == 1)
        good_data &= (xx.wet_pc_10 > self.tcw_threshold)
        good_data &= (xx.pv_pc_10 > self.pv_thresholds[0]) & (xx.qa == 2) | (xx.qa == 1)
       
        notsure_mask = (xx.qa == 1)
    
        cover_type = xx.pv_pc_10.copy(True)
        cover_type.data = dask.array.zeros_like(cover_type.data)
        for s_t in self.pv_thresholds:
            cover_type.data += (xx.pv_pc_10.data > s_t).astype(np.uint8)

        cover_type = erase_bad(cover_type, notsure_mask, nodata=0)
        cover_type = keep_good_only(cover_type, good_data, nodata=NODATA)
        cover_type.attrs['nodata'] = NODATA

        cover_type = cover_type.to_dataset(name="canopy_cover_class")
        # don't want the dimension spec from input but keep the info in case
        if "spec" in cover_type.dims:
            cover_type = cover_type.squeeze(dim=["spec"])
        return cover_type 


register("mangroves", Mangroves)



from random import random
from typing import Optional, Tuple, Union, Callable, Any, Dict, List, Iterable, Iterator
import random
from types import SimpleNamespace
from collections import namedtuple
from datetime import datetime
from itertools import islice
import pickle
import json
import os
from tqdm.auto import tqdm
from urllib.parse import urlparse
import logging
import ciso8601

from odc.dscache import DatasetCache
from datacube import Datacube
from datacube.model import Dataset, GridSpec
from datacube.utils.geometry import Geometry
from datacube.utils.documents import transform_object_tree
from datacube.utils.dates import normalise_dt

from odc.dscache.tools import chopped_dss, bin_dataset_stream, dataset_count, all_datasets
from odc.dscache.tools.tiling import parse_gridspec_with_name
from odc.dscache.tools.profiling import ds_stream_test_func
from ._text import split_and_check

from odc.aws import s3_download, s3_url_parse
from itertools import chain

from .model import DateTimeRange, Task, OutputProduct, TileIdx, TileIdx_txy, TileIdx_xy
from ._gjson import gs_bounds, compute_grid_info, gjson_from_tasks
from .utils import bin_annual, bin_full_history, bin_generic, bin_seasonal

TilesRange2d = Tuple[Tuple[int, int], Tuple[int, int]]
CompressedDataset = namedtuple("CompressedDataset", ["id", "time"])


def _xy(tidx: TileIdx) -> TileIdx_xy:
    return tidx[-2:]


def compress_ds(ds: Dataset) -> CompressedDataset:
    dt = normalise_dt(ds.center_time)
    return CompressedDataset(ds.id, dt)


def is_tile_in(tidx: Tuple[int, int], tiles: TilesRange2d) -> bool:
    (x0, x1), (y0, y1) = tiles
    x, y = tidx
    return (x0 <= x < x1) and (y0 <= y < y1)


def out_path(suffix: str, base: str) -> str:
    if base.endswith(".db"):
        base = base[:-3]
    return base + suffix


def sanitize_query(query):
    def sanitize(v):
        if isinstance(v, Geometry):
            return v.json
        if isinstance(v, datetime):
            return v.isoformat()
        return v

    return transform_object_tree(sanitize, query)


def render_task(tidx: TileIdx_txy) -> str:
    period, xi, yi = tidx
    return f"{period}/{xi:02d}/{yi:02d}"


def parse_task(s: str) -> TileIdx_txy:
    sep = "/" if "/" in s else ","
    t, x, y = split_and_check(s, sep, 3)
    if t.startswith("x"):
        t, x, y = y, t, x
    return (t, int(x.lstrip("x")), int(y.lstrip("y")))

def render_sqs(tidx: TileIdx_txy, filedb: str) -> Dict[str, str]:
    """
    Add extra layer to render task. Convert it to JSON for SQS message body.
    """
    period, xi, yi = tidx
    return {"filedb": filedb, "tile_idx": f"{period}/{xi:02d}/{yi:02d}"}


def parse_sqs(s: str) -> Tuple[TileIdx_txy, str]:
    """
    Add extra layer to parse task. Convert it from JSON for SQS message body.
    """

    message_body = json.loads(s)
    filedb = message_body.get("filedb", None)

    tile_info = message_body.get("tile_idx", None)

    sep = "/" if "/" in s else ","
    t, x, y = split_and_check(tile_info, sep, 3)
    if t.startswith("x"):
        t, x, y = y, t, x
    return ((t, int(x.lstrip("x")), int(y.lstrip("y"))), filedb)


class SaveTasks:
    def __init__(
        self,
        output: str,
        grid: str,
        frequency: str = "annual",
        overwrite: bool = False,
        complevel: int = 6,
    ):

        if DatasetCache.exists(output) and overwrite is False:
            raise ValueError(f"File database already exists: {output}")

        grid, gridspec = parse_gridspec_with_name(grid)

        self._output = output
        self._overwrite = overwrite
        self._complevel = complevel
        self._grid = grid
        self._gridspec = gridspec
        self._frequency = frequency

    def out_path(self, suffix: str) -> str:
        return out_path(suffix, self._output)

    def _get_dss(
        self,
        dc: Datacube,
        product: str,
        msg: Callable[[str], Any],
        dataset_filter: Optional[dict] = {},
        temporal_range: Optional[DateTimeRange] = None,
        tiles: Optional[TilesRange2d] = None,
    ):
        """
        This returns a tuple containing:
        - a generator of datasets
        - the number of datasets in the generator
        - a config dictionary containing the product, temporal range, tiles, and the datacube query used
        """

        cfg: Dict[str, Any] = dict(
            grid=self._grid,
            freq=self._frequency,
        )

        query = dict(product=product, **dataset_filter)

        if tiles is not None:
            (x0, x1), (y0, y1) = tiles
            msg(f"Limit search to tiles: x:[{x0}, {x1}) y:[{y0}, {y1})")
            cfg["tiles"] = tiles
            query["geopolygon"] = gs_bounds(self._gridspec, tiles)

        if temporal_range is not None:
            query.update(
                temporal_range.dc_query(pad=0.6)
            )  # pad a bit more than half a day on each side
            cfg["temporal_range"] = temporal_range.short

        cfg["query"] = sanitize_query(query)

        if DatasetCache.exists(self._output) and self._overwrite is False:
            raise ValueError(f"File database already exists: {self._output}")

        msg("Connecting to the database, counting datasets")
        n_dss = dataset_count(dc.index, **query)
        if n_dss == 0:
            msg("Found no datasets to process")
            return [], n_dss, cfg

        msg(f"Processing {n_dss:,d} datasets")

        if "time" in query:
            dss = chopped_dss(dc, freq="w", **query)
        else:
            if len(query) == 1:
                dss = all_datasets(dc, **query)
            else:
                # note: this blocks for large result sets
                dss = dc.find_datasets_lazy(**query)

        return dss, n_dss, cfg

    def save(
        self,
        dc: Datacube,
        product: str,
        dataset_filter: Optional[dict] = {},
        temporal_range: Union[str, DateTimeRange, None] = None,
        tiles: Optional[TilesRange2d] = None,
        predicate: Optional[Callable[[Dataset], bool]] = None,
        msg: Optional[Callable[[str], Any]] = None,
        debug: bool = False,
        dss=None,
        n_dss=None,
    ) -> bool:
        """
        :param product: Product name to consume
        :param dataset_filter: Optionally apply search filter on Datasets
        :param temporal_range: Optionally  limit query in time
        :param tiles: Optionally limit query to a range of tiles
        :param predicate: If supplied filter Datasets as they come in with custom filter, Dataset->Bool
        :param msg: Observe messages if needed via callback
        :param debug: Dump some intermediate state to files for debugging
        :param dss: A generator of datasets to use
        :param n_dss: The number of datasets in the generator
        """

        dt_range = SimpleNamespace(start=None, end=None)

        def _update_start_end(x, out):
            if out.start is None:
                out.start = x
                out.end = x
            else:
                out.start = min(out.start, x)
                out.end = max(out.end, x)

        def persist(ds: Dataset) -> CompressedDataset:
            _ds = compress_ds(ds)
            _update_start_end(_ds.time, dt_range)
            return _ds

        def msg_default(msg):
            pass

        if msg is None:
            msg = msg_default

        if isinstance(temporal_range, str):
            temporal_range = DateTimeRange(temporal_range)

        if '-' in product:
            products = product.split('-')
            n_dss = 0
            dss = iter(())
            for _product in products:
                dss_found = self._get_dss(dc, _product, msg, dataset_filter, temporal_range, tiles)
                if dss_found:
                    _dss, _n_dss, cfg = dss_found
                    dss = chain.from_iterable((dss, _dss))
                    n_dss += _n_dss
        elif dss is None:
            dss_found = self._get_dss(dc, product, msg, dataset_filter, temporal_range, tiles)
            if dss_found:
                dss, n_dss, cfg = dss_found
        else:

            cfg: Dict[str, Any] = dict(
                grid=self._grid,
                freq=self._frequency,
            )

            if temporal_range is not None:
                cfg["temporal_range"] = temporal_range.short

            if tiles is not None:
                cfg["tiles"] = tiles

        if DatasetCache.exists(self._output) and self._overwrite is False:
            raise ValueError(f"File database already exists: {self._output}")

        if dss is not None and n_dss > 0:
            msg(f"Processing {n_dss:,d} datasets")

            msg("Training compression dictionary")
            dss_slice = list(islice(dss, 0, 100))
            samples = dss_slice.copy()
            random.shuffle(samples)
            zdict = DatasetCache.train_dictionary(samples, 8 * 1024)
            dss = chain(dss_slice, dss)
            msg(".. done")

            cache = DatasetCache.create(
                self._output,
                zdict=zdict,
                complevel=self._complevel,
                truncate=self._overwrite,
            )
            cache.add_grid(self._gridspec, self._grid)
            cache.append_info_dict("stats/", dict(config=cfg))

            cells: Dict[Tuple[int, int], Any] = {}

            if predicate is not None:
                dss = filter(predicate, dss)
            dss = cache.tee(dss)
            dss = bin_dataset_stream(self._gridspec, dss, cells, persist=persist)
            dss = tqdm(dss, total=n_dss)

            rr = ds_stream_test_func(dss)
            msg(rr.text)

            if tiles is not None:
                # prune out tiles that were not requested
                cells = {
                    tidx: cell for tidx, cell in cells.items() if is_tile_in(tidx, tiles)
                }

            if temporal_range is not None:
                # Prune Datasets outside of temporal range (after correcting for UTC offset)
                for cell in cells.values():
                    utc_offset = cell.utc_offset
                    cell.dss = [
                        ds for ds in cell.dss if (ds.time + utc_offset) in temporal_range
                    ]

            n_tiles = len(cells)
            msg(f"Total of {n_tiles:,d} spatial tiles")

            if self._frequency == "all":
                tasks = bin_full_history(cells, start=dt_range.start, end=dt_range.end)
            elif self._frequency == "semiannual":
                tasks = bin_seasonal(cells, months=6, anchor=1)
            elif self._frequency == "seasonal":
                tasks = bin_seasonal(cells, months=3, anchor=12)
            elif self._frequency == "nov-mar":
                tasks = bin_seasonal(cells, months=5, anchor=11, extract_single_season=True)
            elif self._frequency == "apr-oct":
                tasks = bin_seasonal(cells, months=7, anchor=4, extract_single_season=True)
            elif self._frequency == "annual-fy":
                tasks = bin_seasonal(cells, months=12, anchor=7)
            elif self._frequency == "annual":
                tasks = bin_annual(cells)
            elif temporal_range is not None:
                tasks = bin_generic(cells, [temporal_range])
            else:
                tasks = bin_annual(cells)
            # Remove duplicate source uuids.
            # Duplicates occur when queried datasets are captured around UTC midnight
            # and around weekly boundary
            tasks = {k: set(dss) for k, dss in tasks.items()}
            tasks_uuid = {k: [ds.id for ds in dss] for k, dss in tasks.items()}

            all_ids = set()
            for k, dss in tasks_uuid.items():
                all_ids.update(dss)
            msg(f"Total of {len(all_ids):,d} unique dataset IDs after filtering")

            msg(f"Saving tasks to disk ({len(tasks)})")
            cache.add_grid_tiles(self._grid, tasks_uuid)
            msg(".. done")

            self._write_info(tasks, msg, cells, debug)

        return True

    def _write_info(self, tasks, msg, cells, debug):
        csv_path = self.out_path(".csv")
        msg(f"Writing summary to {csv_path}")
        with open(csv_path, "wt") as f:
            f.write('"T","X","Y","datasets","days"\n')

            for p, x, y in sorted(tasks):
                dss = tasks[(p, x, y)]
                n_dss = len(dss)
                n_days = len(set(ds.time.date() for ds in dss))
                line = f'"{p}", {x:+05d}, {y:+05d}, {n_dss:4d}, {n_days:4d}\n'
                f.write(line)

        msg("Dumping GeoJSON(s)")
        grid_info = compute_grid_info(
            cells, resolution=max(self._gridspec.tile_size) / 4
        )
        tasks_geo = gjson_from_tasks(tasks, grid_info)
        for temporal_range, gjson in tasks_geo.items():
            fname = self.out_path(f"-{temporal_range}.geojson")
            msg(f"..writing to {fname}")
            with open(fname, "wt") as f:
                json.dump(gjson, f)

        if debug:
            pkl_path = self.out_path("-cells.pkl")
            msg(f"Saving debug info to: {pkl_path}")
            with open(pkl_path, "wb") as fb:
                pickle.dump(cells, fb)

            pkl_path = self.out_path("-tasks.pkl")
            msg(f"Saving debug info to: {pkl_path}")
            with open(pkl_path, "wb") as fb:
                pickle.dump(tasks, fb)

        return True


class TaskReader:
    def __init__(
        self,
        cache: Union[str, DatasetCache],
        product: Optional[OutputProduct] = None,
        resolution: Optional[Tuple[float, float]] = None
    ):
        self._cache_path = None

        if len(cache) != 0 and isinstance(cache, str):
            if cache.startswith("s3://"):
                self._cache_path = s3_download(cache)
                cache = self._cache_path
            cache = DatasetCache.open_ro(cache)

            # TODO: verify this things are set in the file
            cfg = cache.get_info_dict("stats/config")
            grid = cfg["grid"]
            gridspec = cache.grids[grid]
        else: # if read from message, there is no filedb at beginning
            cfg = {}

        self._product = product
        # if not from_sqs, the resolution check can finish before init TaskReader
        self.resolution = resolution
        self._dscache = cache
        self._cfg = cfg
        self._grid = grid if cache else ""
        self._gridspec = gridspec if cache else ""
        self._all_tiles = sorted(idx for idx, _ in cache.tiles(grid)) if cache else []

    def is_compatible_resolution(self, resolution: Tuple[float, float], tol=1e-8):
        for res, sz in zip(resolution, self._gridspec.tile_size):
            res = abs(res)
            npix = int(sz / res)
            if abs(npix * res - sz) > tol:
                return False
        return True

    def change_resolution(self, resolution: Tuple[float, float]):
        """
        Modify GridSpec to have different pixel resolution but still covering same tiles as the original.
        """
        if not self.is_compatible_resolution(resolution):
            raise ValueError(
                "Supplied resolution is not compatible with the current GridSpec"
            )
        gs = self._gridspec
        self._gridspec = GridSpec(
            gs.crs, gs.tile_size, resolution=resolution, origin=gs.origin
        )

    def init_from_sqs(self, cache: Union[str, DatasetCache]):
        """
        Adding the missing _grid, _gridspec, _gridspec and _all_tiles which skip for sqs task init.
        Upading the cfg which used placeholder filedb path for sqs task init.
        """

        _log = logging.getLogger(__name__)

        self._cache_path = None

        if isinstance(cache, str):
            if cache.startswith("s3://"):
                self._cache_path = s3_download(cache)
                cache = self._cache_path
            cache = DatasetCache.open_ro(cache)

        # TODO: verify this things are set in the file
        cfg = cache.get_info_dict("stats/config")
        grid = cfg["grid"]
        gridspec = cache.grids[grid]

        cfg['filedb'] = cache

        self._dscache = cache
        self._cfg = cfg
        self._grid = grid
        self._gridspec = gridspec
        self._all_tiles = sorted(idx for idx, _ in cache.tiles(grid)) if cache else []

        # first time to access the filedb, then it can do the resolution check
        if self.resolution is not None:
            _log.info(f"Changing resolution to {self.resolution[0], self.resolution[1]}")
            if self.is_compatible_resolution(self.resolution):
                self.change_resolution(self.resolution)
            else: # if resolution has issue, stop init
                _log.error(
                    f"Requested resolution is not compatible with GridSpec in '{cfg.filedb}'"
                )
                raise ValueError(
                    f"Requested resolution is not compatible with GridSpec in '{cfg.filedb}'"
                )


    def __del__(self):
        if self._cache_path is not None:
            os.unlink(self._cache_path)

    def __repr__(self) -> str:
        grid, path, n = self._grid, str(self._dscache.path), len(self._all_tiles)
        return f"<{path}> grid:{grid} n:{n:,d}"

    def _resolve_product(self, product: Optional[OutputProduct]) -> OutputProduct:
        if product is None:
            product = self._product

        if product is None:
            raise ValueError("Product is not supplied and default is not set")
        return product

    @property
    def product(self) -> OutputProduct:
        return self._resolve_product(None)

    @property
    def all_tiles(self) -> List[TileIdx_txy]:
        return self._all_tiles

    def datasets(self, tile_index: TileIdx_txy) -> Tuple[Dataset, ...]:
        return tuple(
            ds for ds in self._dscache.stream_grid_tile(tile_index, self._grid)
        )

    def load_task(
        self,
        tile_index: TileIdx_txy,
        product: Optional[OutputProduct] = None,
        source: Any = None,
        ds_filters: Optional[str] = None,
    ) -> Task:
        product = self._resolve_product(product)

        dss = self.datasets(tile_index)
        if ds_filters is not None:
            ds_checker = DatasetChecker(ds_filters)
            dss = tuple(ds for ds in dss if ds_checker.check_dataset(ds))
        tidx_xy = _xy(tile_index)

        return Task(
            product=product,
            tile_index=tidx_xy,
            geobox=self._gridspec.tile_geobox(tidx_xy),
            time_range=DateTimeRange(tile_index[0]),
            datasets=dss,
            source=source,
        )

    def stream(
        self, tiles: Iterable[TileIdx_txy], product: Optional[OutputProduct] = None, ds_filters: Optional[str] = None,
    ) -> Iterator[Task]:
        product = self._resolve_product(product)
        for tidx in tiles:
            yield self.load_task(tidx, product, ds_filters=ds_filters)

    def stream_from_sqs(
        self,
        sqs_queue,
        product: Optional[OutputProduct] = None,
        visibility_timeout: int = 300,
        ds_filters: Optional[str] = None,
        **kw,
    ) -> Iterator[Task]:
        from odc.aws.queue import get_messages, get_queue
        from ._sqs import SQSWorkToken

        product = self._resolve_product(product)

        if isinstance(sqs_queue, str):
            sqs_queue = get_queue(sqs_queue)

        for msg in get_messages(sqs_queue, visibility_timeout=visibility_timeout, **kw):
            token = SQSWorkToken(msg, visibility_timeout)
            tidx, filedb = parse_sqs(msg.body)

            # avoid the download and update again, how ever, we have to setup an exception to handle the unit test filedb
            if urlparse(filedb).scheme == 's3':
                bucket, key = s3_url_parse(filedb)
            else: # if it is the test_tiles.db, we load it from local
                key = filedb
            local_cache_file = key.split("/")[-1]
            if not os.path.isfile(local_cache_file):  # use the download filedb from S3 as the init context flag
                self.init_from_sqs(filedb)
            yield self.load_task(tidx, product, source=token, ds_filters=ds_filters)


class DatasetChecker:

    def __init__(self, ds_filters):
        ds_filters = ds_filters.split('|')
        self.ds_filters = tuple(json.loads(ds_filter) for ds_filter in ds_filters)

    @staticmethod
    def check_dt(ds_filter, datetime_str):
        time_range = DateTimeRange(ds_filter["datetime"])
        dt = ciso8601.parse_datetime_as_naive(datetime_str)
        return dt in time_range

    def check_ds_1(self, ds_filter, ds):
        valid = True
        for key in ds_filter.keys():
            if key == 'datetime':
                valid &= self.check_dt(ds_filter, ds.metadata_doc["properties"][key])
            else:
                valid &= ds_filter[key] == ds.metadata_doc["properties"][key]

        return valid

    def check_dataset(self, ds):
        valid = False
        for ds_filter in self.ds_filters:
            valid |= self.check_ds_1(ds_filter, ds)

        return valid



import toolz
from typing import Dict, Tuple, List, Any, Callable, Optional
from collections import namedtuple
from datetime import datetime
from .model import DateTimeRange, odc_uuid
from datacube.storage import measurement_paths
from datacube.model import Dataset, DatasetType
from datacube.index.eo3 import prep_eo3


CompressedDataset = namedtuple("CompressedDataset", ["id", "time"])
Cell = Any


def _bin_generic(
    dss: List[CompressedDataset], bins: List[DateTimeRange]
) -> Dict[str, List[CompressedDataset]]:
    """
    Dumb O(NM) implementation, N number of dataset, M number of bins.

    For every bin find all datasets that fall in there, and if not empty keep that bin.
    """
    out: Dict[str, List[CompressedDataset]] = {}
    for b in bins:
        _dss = [ds for ds in dss if ds.time in b]
        if len(_dss) > 0:
            out[b.short] = _dss

    return out


def bin_generic(
    cells: Dict[Tuple[int, int], Cell], bins: List[DateTimeRange]
) -> Dict[Tuple[str, int, int], List[CompressedDataset]]:
    tasks: Dict[Tuple[str, int, int], List[CompressedDataset]] = {}
    for tidx, cell in cells.items():
        _bins = _bin_generic(cell.dss, bins)
        for t, dss in _bins.items():
            tasks[(t,) + tidx] = dss

    return tasks


def bin_seasonal(
    cells: Dict[Tuple[int, int], Cell], months: int, anchor: int, extract_single_season=False
) -> Dict[Tuple[str, int, int], List[CompressedDataset]]:
    # mk_single_season_rules is different from mk_season_rules
    # because the mk_season_rules will split the whole year to 2/3/4 seasons
    # but mk_single_season_rules only extract a single season from the whole year
    if extract_single_season:
        binner = season_binner(mk_single_season_rules(months, anchor))
    else:
        binner = season_binner(mk_season_rules(months, anchor))

    tasks = {}
    for tidx, cell in cells.items():
        utc_offset = cell.utc_offset
        grouped = toolz.groupby(lambda ds: binner(ds.time + utc_offset), cell.dss)

        for temporal_k, dss in grouped.items():
            if temporal_k != "":
                tasks[(temporal_k,) + tidx] = dss

    return tasks


def bin_full_history(
    cells: Dict[Tuple[int, int], Cell], start: datetime, end: datetime
) -> Dict[Tuple[str, int, int], List[CompressedDataset]]:
    duration = end.year - start.year + 1
    temporal_key = (f"{start.year}--P{duration}Y",)
    return {temporal_key + k: cell.dss for k, cell in cells.items()}


def bin_annual(
    cells: Dict[Tuple[int, int], Cell]
) -> Dict[Tuple[str, int, int], List[CompressedDataset]]:
    """
    Annual binning
    :param cells: (x,y) -> Cell(dss: List[CompressedDataset], geobox: GeoBox, idx: Tuple[int, int])
    """
    tasks = {}
    for tidx, cell in cells.items():
        utc_offset = cell.utc_offset
        grouped = toolz.groupby(lambda ds: (ds.time + utc_offset).year, cell.dss)

        for year, dss in grouped.items():
            temporal_k = (f"{year}--P1Y",)
            tasks[temporal_k + tidx] = dss

    return tasks

    
def mk_single_season_rules(months: int, anchor: int) -> Dict[int, str]:
    """
    Construct rules for a each year single season summary
    :param months: Length of season in months can be one of [1, 12]
    :param anchor: Start month of the season [1, 12]
    """
    assert 1 <= months <= 12
    assert 1 <= anchor <= 12

    rules: Dict[int, str] = {}

    start_month = anchor

    # not loop as mk_season_rules(). because it only return
    # single season task cache each run
    for m in range(start_month, start_month + months):
        if m > 12:
            m = m - 12
        if months == 12:
            rules[m] = f"{start_month:02d}--P1Y"
        else:
            rules[m] = f"{start_month:02d}--P{months:d}M"

    return rules


def mk_season_rules(months: int, anchor: int) -> Dict[int, str]:
    """
    Construct rules for a regular seasons
    :param months: Length of season in months can be one of (1,2,3,4,6,12)
    :param anchor: Start month of one of the seasons [1, 12]
    """
    assert months in (1, 2, 3, 4, 6, 12)
    assert 1 <= anchor <= 12

    rules: Dict[int, str] = {}
    for i in range(12 // months):
        start_month = anchor + i * months
        if start_month > 12:
            start_month -= 12

        for m in range(start_month, start_month + months):
            if m > 12:
                m = m - 12
            if months == 12:
                rules[m] = f"{start_month:02d}--P1Y"
            else:
                rules[m] = f"{start_month:02d}--P{months:d}M"

    return rules


def season_binner(rules: Dict[int, str]) -> Callable[[datetime], str]:
    """
    Construct mapping from datetime to a string in the form like 2010-06--P3M

    :param rules: Is a mapping from month (1-Jan, 2-Feb) to a string in the
                  form "{month:int}--P{N:int}M", where ``month`` is a starting
                  month of the season and ``N`` is a duration of the season in
                  months.
    """
    _rules: Dict[int, Tuple[str, int]] = {}

    for month in range(1, 12 + 1):
        season = rules.get(month, "")
        if season == "":
            _rules[month] = ("", 0)
        else:
            start_month = int(season.split("--")[0])
            _rules[month] = (season, 0 if start_month <= month else -1)

    def label(dt: datetime) -> str:
        season, yoffset = _rules[dt.month]
        if season == "":
            return ""
        y = dt.year + yoffset
        return f"{y}-{season}"

    return label


def dedup_s2_datasets(dss):
    """
    De-duplicate Sentinel 2 datasets. Datasets that share same timestamp and
    region code are considered to be duplicates.

    - Sort Datasets by ``(time, region code, label)``
    - Find groups of dataset that share common ``(time, region_code)``
    - Out of duplicate groups pick one with the most recent timestamp in the label (processing time)

    Above, ``label`` is something like this:
    ``S2B_MSIL2A_20190507T093039_N0212_R136_T32NPF_20190507T122231``

    The two timestamps are "capture time" and "processing time".

    :returns: Two list of Datasets, first one contains "freshest" datasets and
              has no duplicates, and the second one contains less fresh duplicates.
    """
    dss = sorted(
        dss,
        key=lambda ds: (
            ds.center_time,
            ds.metadata.region_code,
            ds.metadata_doc["label"],
        ),
    )
    out = []
    skipped = []

    for chunk in toolz.partitionby(
        lambda ds: (ds.center_time, ds.metadata.region_code), dss
    ):
        out.append(chunk[-1])
        skipped.extend(chunk[:-1])
    return out, skipped


def fuse_products(type_1: DatasetType, type_2: DatasetType) -> DatasetType:
    """
    Fuses two products. This function requires access to a Datacube to access the metadata type.

    Fusing two products requires that:
      - both metadata types are eo3
      - there are no conflicting band names
      - the file formats are identical
    """

    def_1, def_2 = type_1.definition, type_2.definition
    fused_def = dict()

    if not def_1["metadata_type"] == def_2["metadata_type"]:
        raise ValueError("metadata_type was different between scenes")

    if not def_1["metadata_type"] == "eo3":
        raise ValueError("metadata_type must be eo3")

    measurements_1 = set(m["name"] for m in def_1["measurements"])
    measurements_2 = set(m["name"] for m in def_2["measurements"])

    if not len(measurements_1.intersection(measurements_2)) == 0:
        raise ValueError("Measurements are overlapping, they should be different")

    file_format = None
    try:
        file_format = def_1["metadata"]["properties"]["odc:file_format"]
        if not file_format == def_2["metadata"]["properties"]["odc:file_format"]:
            raise ValueError("odc:file_format was different between scenes")
    except KeyError:
        # odc:file_format didn't exist, but it's not required, so we're good
        pass

    name = f"fused__{def_1['name']}__{def_2['name']}"

    fused_def["name"] = name
    fused_def["metadata"] = {
        "product": {"name": name}
    }

    if file_format is not None:
        fused_def["metadata"]["odc:file_format"] = file_format

    fused_def[
        "description"
    ] = f"Fused products: {def_1['description']}, {def_2['description']}"
    fused_def["measurements"] = def_1["measurements"] + def_2["measurements"]
    fused_def["metadata_type"] = def_1["metadata_type"]

    return DatasetType(type_1.metadata_type, fused_def)


def fuse_ds(
    ds_1: Dataset, ds_2: Dataset, product: Optional[DatasetType] = None
) -> Dataset:
    """
    This function fuses two datasets. It requires that:
      - the products are fusable
      - grids with the same name are identical
      - labels are in the format 'product_suffix' with identical suffixes
      - CRSs' are identical
      - datetimes are identical
      - $schemas are identical
    """

    doc_1, doc_2 = ds_1.metadata_doc, ds_2.metadata_doc

    if product is None:
        product = fuse_products(ds_1.type, ds_2.type)

    fused_doc = dict()

    fused_doc["id"] = str(
        odc_uuid(product.name, "0.0.0", sources=[doc_1["id"], doc_2["id"]])
    )
    fused_doc["lineage"] = {"source_datasets": [doc_1["id"], doc_2["id"]]}

    # check that all grids with the same name are identical
    common_grids = set(doc_1["grids"].keys()).intersection(doc_2["grids"].keys())
    if not all(doc_1["grids"][g] == doc_2["grids"][g] for g in common_grids):
        raise ValueError("Grids are not all the same")

    # TODO: handle the case that grids have conflicts in a seperate function
    fused_doc["grids"] = {**doc_1["grids"], **doc_2["grids"]}

    # This is currently required, but Alex isn't sure that it should be.
    label_title_doc_1 = doc_1.get("label", toolz.get_in(["properties", "title"], doc_1))
    label_title_doc_2 = doc_2.get("label", toolz.get_in(["properties", "title"], doc_2))

    if label_title_doc_1 is None or label_title_doc_2 is None:
        raise ValueError("No label or title field found found")
    else:
        label_title_doc_1 = label_title_doc_1.replace(doc_1["product"]["name"], "")
        label_title_doc_2 = label_title_doc_2.replace(doc_2["product"]["name"], "")
        if label_title_doc_1 != label_title_doc_2:
            raise ValueError(f"Label/Title field {label_title_doc_1} is not the same as {label_title_doc_2}")

        fused_doc["label"] = f"{product.name}{label_title_doc_1}"

    equal_keys = ["$schema", "crs"]
    for key in equal_keys:
        if not doc_1[key] == doc_2[key]:
            raise ValueError(f"{key} is not the same")
        fused_doc[key] = doc_1[key]

    fused_doc["properties"] = dict()
    # datetime is the only mandatory property
    if not doc_1["properties"]["datetime"] == doc_2["properties"]["datetime"]:
        raise ValueError("Datetimes are not the same")

    # copy over all identical properties
    for key, val in doc_1["properties"].items():
        if val == doc_2["properties"].get(key, None):
            fused_doc["properties"][key] = val

    fused_doc["measurements"] = {**doc_1["measurements"], **doc_2["measurements"]}
    for key, path in {**measurement_paths(ds_1), **measurement_paths(ds_2)}.items():
        fused_doc["measurements"][key]["path"] = path

    fused_ds = Dataset(product, prep_eo3(fused_doc), uris=[""])
    fused_doc["properties"]["fused"] = "True"
    return fused_ds



"""
Water Observations Summaries

Water Observations Summaries are made up of:

- `count_clear`: a count of every time a pixel was observed (not obscured by terrain or clouds)
- `count_wet`: a count of every time a pixel was observed and wet
- `frequency`: what fraction of time (wet/clear) was the pixel wet

The counts are stored as `int16` and the frequency as `float32`.

There are two different Stats Plugin classes implemented in this module. The first generates summary data from
individual water observations, and the second generates a summary of summaries, which is used when generating an all
of time summary from existing annual summaries.

"""
from typing import Sequence, Tuple
import numpy as np
import xarray as xr
from datacube.model import Dataset
from datacube.utils.geometry import GeoBox
from odc.algo import safe_div, apply_numexpr, keep_good_only, binary_dilation
from odc.stac import dc_load
from ._registry import StatsPluginInterface, register


class StatsWofs(StatsPluginInterface):
    """
    Generate a Summary of Water Observations data from individual observations

    The summary is made up of counts of visible and visible and wet, and the frequency of visible and wet.

    Output data types are:
    - `count_clear`: `int16`
    - `count_wet`: `int16`
    - `frequency`: `float32`

    Special care is taken when handling NaN values and `no-data` values.
    """

    NAME = "ga_ls_wo_summary"
    SHORT_NAME = NAME
    VERSION = "1.6.0"
    PRODUCT_FAMILY = "wo_summary"

    # these get padded out if dilation was requested
    BAD_BITS_MASK = 0b0110_1000  # Cloud/Shadow + Terrain Shadow

    def __init__(
        self,
        dilation: int = 0,
        **kwargs
    ):
        super().__init__(input_bands=["water"], **kwargs)
        self._dilation = dilation  # number of pixels to pad around BAD pixels

    @property
    def measurements(self) -> Tuple[str, ...]:
        return "count_wet", "count_clear", "frequency"

    def native_transform(self, xx):
        """
        xx.water -- uint8 classifier bitmask

        .. code-block::

          |128| 64| 32| 16| |  8|  4|  2|  1|
          |---|---|---|---|=|---|---|---|---|
            7   6   5   4     3   2   1   0
            |   |   |   |     |   |   |   |
            |   |   |   |     |   |   |   x---> NODATA: 1 -- all bands were nodata
            |   |   |   |     |   |   o-------> Non Contiguous - some bands were nodata)
            |   |   |   |     |   x-----------> Low Solar Angle
            |   |   |   |     o---------------> Terrain Shadow
            |   |   |   |
            |   |   |   x---------------------> Terrain High Slope
            |   |   o-------------------------> Cloud Shadow
            |   x-----------------------------> Cloud
            o---------------------------------> Water

        out:
          .bad<Bool>   - pixel should not be counted
          .some<Bool>  - there is data (bad or good but not nodata)
          .dry<Bool>   - pixel has dry classification and is not ``bad``
          .wet<Bool>   - pixel has wet classification and is not ``bad``
        """
        if self._dilation != 0:
            xx["bad"] = binary_dilation(
                (xx.water & self.BAD_BITS_MASK) > 0, self._dilation
            ) | ((xx.water & 0b0111_1110) > 0)
        else:
            xx["bad"] = (xx.water & 0b0111_1110) > 0

        # some = (x.water&3)==0, i.e. nodata==0 and non_contigous==0
        xx["some"] = apply_numexpr("((water<<30)>>30)==0", xx, name="some")
        xx["dry"] = xx.water == 0
        xx["wet"] = xx.water == 128
        xx = xx.drop_vars("water")
        for dv in xx.data_vars.values():
            dv.attrs.pop("nodata", None)

        return xx

    @staticmethod
    def fuser(xx):
        """
        xx.bad  -- don't count
        xx.wet  -- is wet
        xx.dry  -- is dry
        xx.some -- there was at least one non-nodata observation at that pixel
        """
        from odc.algo._masking import _or_fuser

        # Merge everything with OR first
        xx = _or_fuser(xx)

        # Ensure all 3 bits are exclusive
        #  bad=T, wet=?, dry=? => (wet'=F  , dry'=F)
        #  bad=F, wet=T, dry=T => (wet'=F  , dry'=F)
        #  else                => (wet'=wet, dry'=dry)
        wet = apply_numexpr("wet & (~dry) & (~bad)", xx, dtype="bool")
        dry = apply_numexpr("dry & (~wet) & (~bad)", xx, dtype="bool")

        return xr.Dataset(dict(wet=wet, dry=dry, bad=xx.bad, some=xx.some))

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        nodata = -999
        count_some = xx.some.sum(axis=0, dtype="int16")
        count_wet = xx.wet.sum(axis=0, dtype="int16")
        count_dry = xx.dry.sum(axis=0, dtype="int16")
        count_clear = count_wet + count_dry
        frequency = safe_div(count_wet, count_clear, dtype="float32")

        count_wet.attrs["nodata"] = nodata
        count_clear.attrs["nodata"] = nodata

        is_ok = count_some > 0
        count_wet = keep_good_only(count_wet, is_ok)
        count_clear = keep_good_only(count_clear, is_ok)

        return xr.Dataset(
            dict(
                count_wet=count_wet,
                count_clear=count_clear,
                frequency=frequency,
            )
        )


register("wofs-summary", StatsWofs)


class StatsWofsFullHistory(StatsPluginInterface):
    """
    Generate a Summary of Water Observations data from existing WO Summaries

    This is useful to turn Annual summary data into all of time summaries.

    Output data is the same as `StatsWofs` produces.
    - `count_clear`: `int16`
    - `count_wet`: `int16`
    - `frequency`: `float32`

    Special care is taken with no-data values, both to pass them through and when calculating the counts and
    frequencies.
    """

    NAME = "ga_ls_wo_fq_myear_3"
    SHORT_NAME = NAME
    VERSION = "1.6.0"
    PRODUCT_FAMILY = "wo_summary"

    def __init__(self, **kwargs):
        super().__init__(input_bands=['count_wet', 'count_clear'])

    @property
    def measurements(self) -> Tuple[str, ...]:
        return "count_wet", "count_clear", "frequency"

    def input_data(self, datasets: Sequence[Dataset], geobox: GeoBox) -> xr.Dataset:
        return dc_load(
            datasets,
            measurements=self.input_bands,
            geobox=geobox,
            chunks={},
        )

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        dtype = xx.count_clear.dtype
        nodata = dtype.type(xx.count_clear.nodata)

        # `missing` is a record of all pixels that were never observed.
        # Store it separately first, then substitute it back in after computing the counts and
        # frequency.
        missing = (xx.count_clear == xx.count_clear.nodata).all(axis=0)
        cc = apply_numexpr(
            "where(count_clear==nodata, 0, count_clear)",
            xx,
            dtype="int16",
            casting="unsafe",
            nodata=nodata,
        ).sum(axis=0, dtype=dtype)

        cw = apply_numexpr(
            "where(count_wet==nodata, 0, count_wet)",
            xx,
            dtype=dtype,
            casting="unsafe",
            nodata=nodata,
        ).sum(axis=0, dtype=dtype)

        _yy = xr.Dataset(dict(cc=cc, cw=cw, missing=missing))

        frequency = apply_numexpr(
            "where(cc==0, _nan, (_1*cw)/(_1*cc))",
            _yy,
            dtype="float32",
            _1=np.float32(1),
            _nan=np.float32("nan"),
        )

        # Finalise the *count* variables by re-inserting the no-data value based on `missing`
        count_clear = apply_numexpr(
            "where(missing, nodata, cc)",
            _yy,
            dtype=dtype,
            nodata=nodata,
            casting="unsafe",
        )
        count_wet = apply_numexpr(
            "where(missing, nodata, cw)",
            _yy,
            dtype=dtype,
            nodata=nodata,
            casting="unsafe",
        )

        count_clear.attrs["nodata"] = int(nodata)
        count_wet.attrs["nodata"] = int(nodata)

        yy = xr.Dataset(
            dict(count_clear=count_clear, count_wet=count_wet, frequency=frequency)
        )
        return yy


register("wofs-summary-fh", StatsWofsFullHistory)



import click
import itertools
import re
import tempfile
import pandas as pd
from os import path
from io import StringIO
from osgeo import gdal, ogr, osr
from ._cli_common import main

def locate_grids(grid_shape, extent_shape, attr_filter):
    ds_grid = ogr.Open(grid_shape)
    lyr_grid = ds_grid.GetLayer(0)
    lyr_grid.ResetReading()
    grid_spatial_ref = lyr_grid.GetSpatialRef()

    ds_extent = ogr.Open(extent_shape)
    lyr_extent = ds_extent.GetLayer(0)
    if attr_filter is not None:
        lyr_extent.SetAttributeFilter(attr_filter)
    lyr_extent.ResetReading()

    extent_spatial_ref = lyr_extent.GetSpatialRef()
    transform = osr.CoordinateTransformation(grid_spatial_ref, extent_spatial_ref)
    extent_grids = []
    for grid in lyr_grid:
        grid_geom = grid.geometry()
        grid_geom.Transform(transform)
        lyr_extent.SetSpatialFilter(grid_geom)
        if lyr_extent.GetFeatureCount() > 0:
            extent_grids += [re.findall(r'\d+', grid['region_code'])]
        lyr_extent.SetSpatialFilter(None)
        lyr_extent.ResetReading()
    return extent_grids

@main.command("locate-grids")
@click.option("--attr-filter", type=str, default=None, help="Filter the input shape by attributes, e.g., FEAT_CODE != 'sea'")
@click.argument("grid-shape", type=str)
@click.argument("extent-shape", type=str)
@click.argument("csv-path", type=str, required=False)
@click.option("--verbose", "-v", is_flag=True, help="Be verbose")
def cli(attr_filter, grid_shape, extent_shape, csv_path, verbose):
    """
    Generate a list of grids overlapping with the input shape extent
    
    GRID_SHAPE is the geojson or ESRI shape file of the grids.
    
    EXTENT_SHAPE is the ESRI shape file where the extent covers the grids.
    
    CSV_PATH is the path where the csv of the grids list will be saved, default is None.
    By default, the file will be saved in the system temporary folder.
    
    """

    print("It takes time, not frozen...")
    print("Input shape %s filtered by %s" % (extent_shape, attr_filter))
    extent_grids = locate_grids(grid_shape, extent_shape, attr_filter)
    csv_buffer = StringIO()
    pd.DataFrame(extent_grids).to_csv(csv_buffer, index=None, header=None)
    csv_buffer.seek(0)
    
    if csv_path is None:
        tmp_path = tempfile.gettempdir()
        csv_path = path.join(tmp_path, "extent_grids.csv")
    with open(csv_path, 'w') as f:
        f.write(csv_buffer.read())
    print("Results saved to", csv_path)

if __name__ == "__main__":
    cli()



import sys
import click
from ._cli_common import main, setup_logging, click_resolution, click_yaml_cfg


@main.command("run")
@click.option("--threads", type=int, help="Number of worker threads")
@click.option("--memory-limit", type=str, help="Limit memory used by Dask cluster")
@click.option(
    "--dryrun",
    is_flag=True,
    help="Do not run computation just print what work will be done",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=None,
    help="Do not check if output already exists",
)
@click.option(
    "--heartbeat-filepath",
    type=str,
    help="Path to store pod's heartbeats when running stats as K8 jobs",
)
@click.option(
    "--dataset-filters",
    type=str,
    help="",
)
@click.option(
    "--public/--no-public",
    is_flag=True,
    default=None,
    help="Mark outputs for public access (default: no)",
)
@click.option(
    "--apply_eodatasets3",
    is_flag=True,
    default=False,
    help="Apply eodatasets3 plugin to generate metadata files (default: use PySTAC to generate metadata files)",
)
@click.option(
    "--location", type=str, help="Output location prefix as a uri: s3://bucket/path/"
)
@click.option("--max-processing-time", type=int, help="Max seconds per task")
@click.option("--from-sqs", type=str, help="Read tasks from SQS", default="")
@click_yaml_cfg("--config", help="Runner Config")
@click.option(
    "--plugin", type=str, help="Which stats plugin to run",
)
@click_yaml_cfg(
    "--plugin-config", help="Config for plugin in yaml format, file or text"
)
@click_yaml_cfg("--cog-config", help="Configure COG options")
@click.option("--resampling", type=str, help="Input resampling strategy, e.g. average")
@click_resolution("--resolution", help="Override output resolution")
@click.argument("filedb", type=str, nargs=1, default="")
@click.argument("tasks", type=str, nargs=-1)
def run(
    filedb,
    tasks,
    from_sqs,
    config,
    plugin_config,
    cog_config,
    resampling,
    resolution,
    plugin,
    dryrun,
    threads,
    memory_limit,
    overwrite,
    public,
    location,
    max_processing_time,
    heartbeat_filepath,
    dataset_filters,
    apply_eodatasets3,
):
    """
    Run Stats.

    Task could be one of the 3 things

    \b
    1. Comma-separated triplet: period,x,y or 'x[+-]<int>/y[+-]<int>/period
       2019--P1Y,+003,-004
       2019--P1Y/3/-4          `/` is also accepted
       x+003/y-004/2019--P1Y   is accepted as well
    2. A zero based index
    3. A slice following python convention <start>:<stop>[:<step]
        ::10 -- every tenth task: 0,10,20,..
       1::10 -- every tenth but skip first one 1, 11, 21 ..
        :100 -- first 100 tasks

    If no tasks are supplied and --from-sqs is not used, the whole file will be processed.
    """
    setup_logging()

    import logging
    from .model import TaskRunnerConfig
    from .proc import TaskRunner
    from .plugins import import_all

    _log = logging.getLogger(__name__)

    if from_sqs:
        if dryrun:
            print("Can not dry run from SQS")
            sys.exit(1)
        if len(tasks):
            print("Supply either <tasks> or --from-sqs")
            sys.exit(2)

    import_all()

    if config is None:
        config = {}

    _cfg = dict(**config)
    s3_acl = 'public-read' if public else None

    cfg_from_cli = {
        k: v
        for k, v in dict(
            filedb=filedb,
            plugin=plugin,
            threads=threads,
            memory_limit=memory_limit,
            output_location=location,
            s3_acl=s3_acl,
            overwrite=overwrite,
            max_processing_time=max_processing_time,
            heartbeat_filepath=heartbeat_filepath,
        ).items()
        if v is not None and v != ""
    }

    _log.info(f"Config overrides: {cfg_from_cli}")

    _cfg.update(cfg_from_cli)
    if plugin_config is not None:
        _cfg["plugin_config"] = plugin_config

    if resampling is not None and len(resampling) > 0:
        _cfg.setdefault("plugin_config", {})["resampling"] = resampling

    if cog_config is not None:
        _cfg["cog_opts"] = cog_config

    if from_sqs: # if config or CLI has filedb, but run from sqs, throw this warning message.
        _log.warning("The `filedb` from config or CLI will be a placeholder value. Actual filedb saved in SQS message")
    elif not _cfg.get('filedb'):
        _log.error("Must supply `filedb` either through config or CLI")
        sys.exit(1)    

    
    tmp = _cfg.pop('dataset-filters', None)
    if dataset_filters is None:
        dataset_filters = tmp

    tmp = _cfg.pop('apply_eodatasets3', False)
    if not apply_eodatasets3:
        apply_eodatasets3 = tmp

    cfg = TaskRunnerConfig(**_cfg)
    _log.info(f"Using this config: {cfg}")

    # Warnning, if run from sqs, the runner will skil rdr init
    runner = TaskRunner(cfg, resolution=resolution, from_sqs=from_sqs)

    if dryrun:
        check_exists = runner.verify_setup()
        for task in runner.dry_run(tasks, check_exists=check_exists, ds_filters=dataset_filters):
            print(task.meta)
        sys.exit(0)

    if not runner.verify_setup():
        print("Failed to verify setup, exiting")
        sys.exit(1)

    result_stream = runner.run(sqs=from_sqs, ds_filters=dataset_filters, apply_eodatasets3=apply_eodatasets3) if from_sqs else runner.run(tasks=tasks, ds_filters=dataset_filters, apply_eodatasets3=apply_eodatasets3)

    total = 0
    finished = 0
    skipped = 0
    errored = 0
    for result in result_stream:
        total += 1
        task = result.task
        if result:
            if result.skipped:
                skipped += 1
                _log.info(f"Skipped task #{total:,d}: {task.location} {task.uuid}")
            else:
                finished += 1
                _log.info(f"Finished task #{total:,d}: {task.location} {task.uuid}")
        else:
            errored += 1
            _log.error(f"Failed task #{total:,d}: {task.location} {task.uuid}")

        _log.info(f"T:{total:,d}, OK:{finished:,d}, S:{skipped:,d}, E:{errored:,d}")

    _log.info(
        f"Completed processing {total:,d} tasks, OK:{finished:,d}, S:{skipped:,d}, E:{errored:,d}"
    )

    _log.info("Shutting down Dask cluster")
    del runner
    _log.info("Calling sys.exit(0)")
    sys.exit(0)



"""
Geomedian
"""
from typing import Optional, Mapping, Sequence, Tuple, Iterable
import xarray as xr
from datacube.model import Dataset
from datacube.utils.geometry import GeoBox
from odc.algo import erase_bad, geomedian_with_mads
from odc.algo.io import load_enum_filtered
from ._registry import StatsPluginInterface, register


class StatsGM(StatsPluginInterface):
    NAME = "gm"
    SHORT_NAME = NAME
    VERSION = "0.0.0"
    PRODUCT_FAMILY = "geomedian"

    def __init__(
        self,
        bands: Tuple[str, ...],
        mask_band: str,
        cloud_classes: Tuple[str, ...],
        nodata_classes: Optional[Tuple[str, ...]] = None,
        filters: Optional[Iterable[Tuple[str, int]]] = None,
        basis_band=None,
        aux_names=dict(smad="smad", emad="emad", bcmad="bcmad", count="count"),
        work_chunks: Tuple[int, int] = (400, 400),
        **kwargs,
    ):
        self.bands = tuple(bands)
        self._mask_band = mask_band
        if nodata_classes is not None:
            nodata_classes = tuple(nodata_classes)
        self._nodata_classes = nodata_classes
        input_bands = self.bands
        if self._nodata_classes is not None:
            # NOTE: this ends up loading Mask band twice, once to compute
            # ``.erase`` band and once to compute ``nodata`` mask.
            input_bands = (*input_bands, self._mask_band)
        super().__init__(
            input_bands=input_bands,
            basis=basis_band or self.bands[0],
            **kwargs)

        self.cloud_classes = tuple(cloud_classes)
        # if filters:
        #     self.filters: Optional[Mapping] = dict(filters)
        self.filters = filters
        # else:
        #     self.filters = None

        self._renames = aux_names
        self.aux_bands = tuple(
            self._renames.get(k, k) for k in ("smad", "emad", "bcmad", "count")
        )

        self._work_chunks = work_chunks

    @property
    def measurements(self) -> Tuple[str, ...]:
        return self.bands + self.aux_bands

    def native_transform(self, xx: xr.Dataset) -> xr.Dataset:
        from odc.algo import enum_to_bool, keep_good_only

        if not self._mask_band in xx.data_vars:
            return xx

        # Erase Data Pixels for which mask == nodata
        #
        #  xx[mask == nodata] = nodata
        mask = xx[self._mask_band]
        xx = xx.drop_vars([self._mask_band])
        keeps = enum_to_bool(mask, self._nodata_classes, invert=True)
        xx = keep_good_only(xx, keeps)

        return xx

    def input_data(self, datasets: Sequence[Dataset], geobox: GeoBox) -> xr.Dataset:
        erased = load_enum_filtered(
            datasets,
            self._mask_band,
            geobox,
            categories=self.cloud_classes,
            filters=self.filters,
            groupby=self.group_by,
            resampling=self.resampling,
            chunks={},
        )
        xx = super().input_data(datasets, geobox)
        xx = erase_bad(xx, erased)
        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        scale = 1 / 10_000
        cfg = dict(
            maxiters=1000,
            num_threads=1,
            scale=scale,
            offset=-1 * scale,
            reshape_strategy="mem",
            out_chunks=(-1, -1, -1),
            work_chunks=self._work_chunks,
            compute_count=True,
            compute_mads=True,
        )

        gm = geomedian_with_mads(xx, **cfg)
        gm = gm.rename(self._renames)

        return gm


register("gm-generic", StatsGM)


class StatsGMS2(StatsGM):
    NAME = "gm_s2_annual"
    SHORT_NAME = NAME
    VERSION = "0.0.0"
    PRODUCT_FAMILY = "geomedian"

    def __init__(
        self,
        bands: Optional[Tuple[str, ...]] = None,
        mask_band: str = "SCL",
        cloud_classes: Tuple[str, ...] = (
            "cloud shadows",
            "cloud medium probability",
            "cloud high probability",
            "thin cirrus",
        ),
        filters: Optional[Iterable[Tuple[str, int]]] = [("opening", 2), ("dilation",5)],
        aux_names=dict(smad="SMAD", emad="EMAD", bcmad="BCMAD", count="COUNT"),
        rgb_bands=None,
        **kwargs
    ):
        if bands is None:
            bands = (
                "B02",
                "B03",
                "B04",
                "B05",
                "B06",
                "B07",
                "B08",
                "B8A",
                "B11",
                "B12",
            )
            if rgb_bands is None:
                rgb_bands = ("B04", "B03", "B02")

        super().__init__(
            bands=bands,
            mask_band=mask_band,
            cloud_classes=cloud_classes,
            filters=filters,
            aux_names=aux_names,
            rgb_bands=rgb_bands,
            **kwargs
        )


register("gm-s2", StatsGMS2)


class StatsGMLS(StatsGM):
    NAME = "gm_ls_annual"
    SHORT_NAME = NAME
    VERSION = "3.0.0"
    PRODUCT_FAMILY = "geomedian"

    def __init__(
        self,
        bands: Optional[Tuple[str, ...]] = None,
        mask_band: str = "fmask",
        cloud_classes: Tuple[str, ...] = ("cloud", "shadow"),
        nodata_classes: Optional[Tuple[str, ...]] = ("nodata",),
        aux_names=dict(smad="sdev", emad="edev", bcmad="bcdev", count="count"),
        rgb_bands=None,
        **kwargs
    ):
        if bands is None:
            bands = (
                "red",
                "green",
                "blue",
                "nir",
                "swir1",
                "swir2",
            )
            if rgb_bands is None:
                rgb_bands = ("red", "green", "blue")

        super().__init__(
            bands=bands,
            mask_band=mask_band,
            cloud_classes=cloud_classes,
            nodata_classes=nodata_classes,
            aux_names=aux_names,
            rgb_bands=rgb_bands,
            **kwargs,
        )


register("gm-ls", StatsGMLS)



import click
from datacube import Datacube
from odc.algo import save_cog

from ._cli_common import main
from ._stac_fetch import s3_fetch_dss


@main.command("generate-mosaic")
@click.argument("input_glob", type=str)
@click.argument("output", type=str)
@click.option("--bands", type=str)
@click.option("--resampling", type=str, default="average")
@click.option("--verbose", "-v", is_flag=True, help="Be verbose")
def generate_mosaic(input_glob, output, bands, resampling, verbose):
    """
    Generate mosaic overviews of the stats data.

    Note: The input bucket must be public otherwise the data can not be listed.
    """
    bands = bands.split(",")

    if verbose:
        print(f"Preparing mosaics from: '{input_glob}'")

    dss = s3_fetch_dss(input_glob)

    dc = Datacube()
    xx = dc.load(
        datasets=dss,
        dask_chunks={"x": 2048, "y": 2048},
        resolution=(-120, 120),
        measurements=bands,
    )

    if verbose:
        print(f"Writing {output}")

    xx = xx.squeeze("time").to_stacked_array("bands", ["x", "y"])
    yy = save_cog(
        xx,
        output,
        blocksize=1024,
        compress="zstd",
        zstd_level=4,
        overview_resampling=resampling,
        NUM_THREADS="ALL_CPUS",
        bigtiff="YES",
        SPARSE_OK=True,
    )

    yy.compute()


if __name__ == "__main__":
    main()



import math
from copy import deepcopy
import toolz
from typing import Tuple, Dict, Any
from datetime import timedelta

from datacube.model import GridSpec
from datacube.utils.geometry import polygon_from_transform, Geometry
from odc.dscache.tools import solar_offset
from .model import TileIdx_xy, TileIdx_txy


def gs_bounds(gs: GridSpec, tiles: Tuple[Tuple[int, int], Tuple[int, int]]) -> Geometry:
    """
    Compute Polygon for a selection of tiles.

    :param gs: GridSpec
    :param tiles: (x_range, y_range)

    X,Y ranges are inclusive on the left and exclusive on the right, same as numpy slicing.
    """
    ((x0, x1), (y0, y1)) = tiles
    if gs.resolution[0] < 0:
        gb = gs.tile_geobox((x0, y1 - 1))
    else:
        gb = gs.tile_geobox((x0, y0))

    nx = (x1 - x0) * gb.shape[1]
    ny = (y1 - y0) * gb.shape[0]
    return polygon_from_transform(nx, ny, gb.affine, gb.crs)


def timedelta_to_hours(td: timedelta) -> float:
    return td.days * 24 + td.seconds / 3600


def compute_grid_info(
    cells: Dict[TileIdx_xy, Any], resolution: float = math.inf, title_width: int = 0
) -> Dict[TileIdx_xy, Any]:
    """
    Compute geojson feature for every cell in ``cells``.
    Where ``cells`` is produced by ``bin_dataset_stream``
    """
    if title_width == 0:
        nmax = max([max(abs(ix), abs(iy)) for ix, iy in cells])
        # title_width is the number of digits in the index
        title_width = len(str(nmax))

    grid_info = {}

    for idx, cell in cells.items():
        geobox = cell.geobox
        utc_offset = timedelta_to_hours(solar_offset(geobox.extent))
        wrapdateline = utc_offset <= -11 or utc_offset >= 11

        geom = geobox.extent.to_crs(
            "epsg:4326", resolution=resolution, wrapdateline=wrapdateline
        )
        ix, iy = idx
        grid_info[idx] = {
            "type": "Feature",
            "geometry": geom.json,
            "properties": {
                "title": f"{ix:0{title_width}d},{iy:0{title_width}d}",
                "region_code": f"x{ix:0{title_width}d}y{iy:0{title_width}d}",
                "ix": ix,
                "iy": iy,
                "utc_offset": utc_offset,
                "total": len(cell.dss),
            },
        }

    return grid_info


def gjson_from_tasks(
    tasks: Dict[TileIdx_txy, Any], grid_info: Dict[TileIdx_xy, Any]
) -> Dict[str, Dict[str, Any]]:
    """
    Group tasks by time period and compute geosjon describing every tile covered by each time period.

    Returns time_period -> GeoJSON mapping

    Each feature in GeoJSON describes one tile and has following propreties

      .total      -- number of datasets
      .days       -- number of days with at least one observation
      .utc_offset -- utc_offset used to compute day boundaries
    """

    def _get(idx):
        xy_idx = idx[-2:]
        geo = deepcopy(grid_info[xy_idx])

        dss = tasks[idx]
        utc_offset = timedelta(hours=geo["properties"]["utc_offset"])

        ndays = len(set((ds.time + utc_offset).date() for ds in dss))
        geo["properties"]["total"] = len(dss)
        geo["properties"]["days"] = ndays

        return geo

    def process(idxs):
        return dict(type="FeatureCollection", features=[_get(idx) for idx in idxs])

    return {
        t: process(idxs)
        for t, idxs in toolz.groupby(toolz.first, sorted(tasks)).items()
    }



"""
Fractional Cover Percentiles
"""
from functools import partial
from itertools import product
from typing import Optional, Tuple

import numpy as np
import xarray as xr
from odc.algo import keep_good_only
from odc.algo._masking import _fuse_mean_np, _fuse_or_np, _or_fuser, _xr_fuse
from odc.algo._percentile import xr_quantile_bands

from ._registry import StatsPluginInterface, register

NODATA = 255


class StatsFCP(StatsPluginInterface):

    NAME = "ga_fc_percentiles"
    SHORT_NAME = NAME
    VERSION = "0.0.2"
    PRODUCT_FAMILY = "fc_percentiles"

    def __init__(
        self,
        max_sum_limit: Optional[int] = None,
        clip_range: Optional[Tuple] = None,
        ue_threshold: Optional[int] = None,
        count_valid: Optional[bool] = False,
        **kwargs,
    ):
        super().__init__(input_bands=["water", "pv", "bs", "npv", "ue"], **kwargs)

        self.max_sum_limit = max_sum_limit
        self.clip_range = clip_range
        self.ue_threshold = ue_threshold
        self.count_valid = count_valid

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = [
            f"{b}_pc_{p}" for b, p in product(["pv", "bs", "npv"], ["10", "50", "90"])
        ]
        _measurements.append("qa")
        if self.count_valid:
            _measurements.append("count_valid")
        return _measurements

    def native_transform(self, xx):
        """
        Loads data in its native projection. It performs the following:

        1. Load all fc and WOfS bands
        2. Set the high terrain slope flag to 0
        3. Set all pixels that are not clear and dry to NODATA
        4. Calculate the clear wet pixels
        5. Drop the WOfS band
        """

        water = xx.water & 0b1110_1111
        xx = xx.drop_vars(["water"])

        # Pick out the dry pixels
        dry = water == 0

        # Incrementally add to the valid band
        valid = dry

        # Pick out the pixels that have an unmixing error of less than the threshold
        if self.ue_threshold is not None:
            # No QA
            unmixing_error_lt_30 = xx.ue < self.ue_threshold
            valid = valid & unmixing_error_lt_30
        xx = xx.drop_vars(["ue"])

        # If there's a sum limit or clip range, implement these
        if self.max_sum_limit is not None or self.clip_range is not None:
            sum_bands = None
            for band in xx.data_vars.keys():
                attributes = xx[band].attrs
                mask = xx[band] == NODATA
                band_data = keep_good_only(xx[band], ~mask, nodata=0)

                if self.max_sum_limit is not None:
                    if sum_bands is None:
                        sum_bands = band_data
                    else:
                        sum_bands = sum_bands + band_data

                if self.clip_range is not None:
                    # No QA
                    limit_min, limit_max = self.clip_range
                    clipped = np.clip(xx[band], limit_min, limit_max)
                    # Set masked values back to NODATA
                    xx[band] = clipped.where(~mask, NODATA)
                    xx[band].attrs = attributes

            if self.max_sum_limit is not None:
                sum_lt_limit = sum_bands < self.max_sum_limit
                valid = valid & sum_lt_limit

        xx = keep_good_only(xx, valid, nodata=NODATA)

        xx["wet"] = water == 128
        xx["valid"] = valid

        return xx

    def fuser(self, xx):
        wet = xx["wet"]
        valid = xx["valid"]

        xx = _xr_fuse(
            xx.drop_vars(["wet", "valid"]), partial(_fuse_mean_np, nodata=NODATA), ""
        )

        xx["wet"] = _xr_fuse(wet, _fuse_or_np, wet.name)
        xx["valid"] = _xr_fuse(valid, _fuse_or_np, valid.name)

        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        # (!all_bands_valid) & is_ever_wet => 0
        # (!all_bands_valid) & (!is_ever_wet) => 1
        # all_bands_valid => 2

        wet = xx["wet"]
        valid = xx["valid"]
        xx = xx.drop_vars(["wet", "valid"])

        yy = xr_quantile_bands(xx, [0.1, 0.5, 0.9], nodata=NODATA)
        is_ever_wet = _or_fuser(wet).squeeze(wet.dims[0], drop=True)

        band, *bands = yy.data_vars.keys()
        all_bands_valid = yy[band] != NODATA
        for band in bands:
            all_bands_valid &= yy[band] != NODATA

        all_bands_valid = all_bands_valid.astype(np.uint8)
        is_ever_wet = is_ever_wet.astype(np.uint8)
        yy["qa"] = 1 + all_bands_valid - is_ever_wet * (1 - all_bands_valid)

        if self.count_valid:
            yy["count_valid"] = valid.sum(axis=0, dtype="int16")

        return yy


register("fc-percentiles", StatsFCP)
